/* **************************************************************** *
 * Implmentation du Quicksort One-way                              *
 *                                                                  *
 * D. Taralla, 2e Bac. Info.                                        *
 * Last review: 03/03/2011                                          *
 * **************************************************************** */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

/* Cette fonction affiche simplement le contenu actuel d'un tableau sur la 
 * sortie standard. */
static void printArray(int* tab, int size) {
    printf("Actual state:\n");
    for (int i = 0; i < size; i++)
        printf("%d ", tab[i]);
    
    printf("\n");
}

/* Cette fonction Quicksort n'avance que dans un sens : de la gauche vers la 
 * droite comme l'indique son nom. Elle s'articule autour d'une boucle 
 * principale. L'invariant suivant est respect  chaque tour de boucle :
 *     tab[0:l-1] = lments tris et leurs signes sont corrects.
 *     Les lments ngatifs du tableau sont  leur position finale.
 * 
 * Au dpart, le marqueur low pointe vers l'indice 0 et tous les lments sont 
 * positifs sauf le dernier qui vaut -1 (prcondition). 
*/
static void quicksort(int* tab, int low, int high) {
    int i, j;
    int K;

    while (low < high) {
        /* Si le premier lment de la partie qu'on trie est ngatif, il est  
         * la bonne place, on peut lui rendre son signe. Tout ce qui se trouve 
           avant le marqueur low est tri dsormais. */
        if (tab[low] < 0) {
            tab[low] = -tab[low];
            low++;
        }
        // Sinon, trier  partir de low.
        else {
            /* On va trouver la place de tab[low]. On l'enregistre pour pouvoir
             * le mettre  sa place  la fin de l'algorithme. */
            i = j = low;
            K = tab[low];
            inc: j++;

            /* Tant que notre lment K est infrieur ou gal  l'lment de 
             * l'indice j, on incrmente ce dernier car a veut dire que K est 
             * dj bien plac par rapport  ces lments. Mais attention, si 
             * K est bien plac par rapport aux trois lments suivants mais 
             * pas par rapport au quatrime, une manipulation est ncessaire !
             */
            while (K <= tab[j]) j++;

            /* Arriv ici, cela signifie qu'on a trouv un lment qui est mal 
             * tri par rapport  K.
             * Si cet lment est ngatif, cela signifie que celui-ci est  sa 
             * place finale : par contruction de la boucle et par l'invariant, 
             * on sort de la boucle induite par le goto*.
             * Tant qu'on a pas trouv d'lment ngatif (de gauche  droite), 
             * on continue de repositionner et ddoubler. */
            if (tab[j] > 0) {
                tab[i] = tab[j];
                i++;
                tab[j] = tab[i];

                goto inc; // berk, mais ncessaire
            }

            /* L'lment  la position i, c'est notre cl ! On l'y place, puis
             * en prenant son oppos, pour dire qu'elle est  sa place finale
             */
            tab[i] = -K;

            /* L'volution du marqueur i
             * ==========================
             * Au dpart, i vaut low qui vaut 0. i va servir  dterminer la 
             * position finale de K. On incrmente i seulement lorsque l'on a 
             * trouv un tab[j] positif et strictement suprieur  K. Dans le 
             * corps du if, on remarque que l'on ddouble  chaque fois un 
             * lment : tab[i] devient tab[j] et tab[j] devient tab[i+1]. Et 
             * voici le rle du marqueur i : garder la position de l'origine 
             * du dernier doublon cr de cette manire. Ainsi, en fin de 
             * parcours, lorsqu'on n'entre plus dans le if, i pointe vers la 
             * position finale de K. On place -K l, et on retrouve une 
             * permutation du tableau de dpart, sans doublons. */
        }
    }
}

int main() {
    const int N = 21; // 20 + 1 pour le "-1" additionnel


    printf("This program sorts lists of %d positive non-zero integers, from "
           "1 to 99, using 1-way Quicksort.\nIt was NOT made to be robust, "
           "wrong inputs aren't managed!\n", N-1);
    while (true) {
        int seed = 0;

        printf("\nNew test -- Enter random seed (integer): ");
        scanf("%d", &seed);

        srand(seed);
        int a[N];
        for (int i = 0; i < N-1; i++) {
            do a[i] = rand() % 100;
            while (a[i] == 0);
        }

        a[N-1] = -1; // Obligatoire pour cet algorithme

        printArray(a, N-1); // Ne pas afficher le -1 final
        quicksort(a, 0, N);
        printArray(a, N-1); // Ne pas afficher le -1 final
    }

    return 0;
}
