/* **************************************************************** *
 * Implmentation du Quicksort Three-way                            *
 *                                                                  *
 * D. Taralla, 2e Bac. Info.                                        *
 * Last review: 03/03/2011                                          *
 * **************************************************************** */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

struct ret_t {
  int low;
  int mid;
};

typedef struct ret_t ret;

/* Cette fonction affiche simplement le contenu actuel d'un tableau sur la 
 * sortie standard. */
static void printArray(int* tab, int size) {
    printf("Actual state:\n");
    for (int i = 0; i < size; i++)
        printf("%d ", tab[i]);
    
    printf("\n");
}

/* Cette fonction partitionne la partie [low:high] du tableau tab en utilisant 
 * la mthode du Three-way partionning. Elle retourne la position du pivot.
 *
 * Explication dtaille de la mthode:
 * -------------------------------------
 *
 * Cette technique rorganise le tableau de la manire suivante :
 * On observe les lments compris entre 0 et ref - 1.
 * On va diviser le tableau tab[low, ..., ref-1] en 3 parties :  gauche,
 * tous les lments < que tab[ref], suivis des lments gaux  tab[ref] et 
 * ensuite tous les lments > que tab[ref].Quand cette action est effectue,
 * il ne reste plus qu' changer tab[ref] et le tab[mid final].
 *
 * Comment se passe le partitionnement ? A chaque tour de boucle,
 * l'invariant suivant est respect :
 *     tab[0 : low-1]      : < tab[ref] (partie A)
 *     tab[low : mid-1]    : == tab[ref] (partie B)
 *     tab[mid : high]     : indtermin,  analyser (partie C)
 *     tab[high+1 : ref-1] : > tab[ref] (partie D)
 *  
 * La zone indtermine est rduite  chaque tour de boucle, que ce soit par 
 * la droite (dcrment du marqueur high) ou par la gauche (incrment des 
 * marqueurs low et mid). A chaque tour, on regarde si tab[mid] est plus 
 * petit, gal ou suprieur  la rfrence.
 * S'il est infrieur on switche cet lment avec tab[low]. En effet, vu l'
 * invariant, on sait que tous les lments d'indice infrieur stric.  low 
 * sont infrieurs  la rfrence. Donc si ce qu'on vient d'analyser est 
 * lui aussi infrieur  low, cela signifie que sa place est dans la partie
 * A. On agrandit donc A en plaant cet lment en low. La partie A a t
 * largie ; on doit incrmenter low. On incrmente aussi mid car on a 
 * termin l'analyse de cet lment. Rappelons que selon l'invariant
 * toujours, les lments tab[low : mid-1] sont des lments gaux  tab[ref]
 * donc switcher tab[low] et tab[mid] revient  dcaler la partie B de une
 * unit vers la droite. L'invariant est conserv.
 *
 * S'il est gal, on n'y touche pas et on
 * incrmente mid (en effet, vu l'invariant, cela veut dire que l'lment
 * est  sa place : on largit la partie B et on rduit C). L'invariant est
 * toujours respect.
 *
 * Si il est suprieur, on le switche avec l'lment  l'indice high (on
 * ne sait pas ce que celui-ci vaut, btw), on dcrmente ensuite high sans 
 * toucher  mid qui au prochain tour de boucle, permettera d'valuer cet
 * lment switch inconnu et le placer  la bonne place. On sait que l'
 * invariant reste respect : D est largie, alors que C est rduite, et l'
 * lment est bien plac selon l'invariant.
 *
 * A la fin de l'algorithme,  la sortie de la boucle, on place l'lment de
 * rfrence au bon endroit (on le switch avec tab[mid]). Pourquoi mid ?
 * Parce que, que l'on termine la boucle par un high-- ou un mid++, mid
 * pointera vers le premier lment de tab qui est strictement plus grand
 * que tab[ref]. Lors de l'avant-dernire itration, on a high - mid = 1.
 * Ensuite, qu'on excute high-- ou mid++, on arrive  mid = high. Le dernier
 * tour de boucle compare donc tab[mid=high] avec lui-mme, on incrmente
 * donc mid une dernire fois. Dsormais, mid - high = 1, la boucle s'arrte,
 * et on a plus qu' changer tab[ref] avec tab[mid] pour placer tab[ref] au
 * bon endroit.
 * La postcondition est respecte,  savoir, on a dsormais 3 parties (dont
 * l'une peut tre vide) respectant l'invariant. */
static ret* partition(int* tab, int low, int high) {
    int ref = high--; // Mmoriser l'indice de l'lment de rfrence
    int mid = low;

    // Partionnement
    while (mid <= high) {
        if (tab[mid] < tab[ref]) {
            int tmp  = tab[low];
            tab[low] = tab[mid];
            tab[mid] = tmp;
            low++;
            mid++;
        }
        else if (tab[mid] == tab[ref]) {
            mid++;
        }
        else {
            int tmp   = tab[high];
            tab[high] = tab[mid];
            tab[mid]  = tmp;
            high--;
        }
    }

    // Mettre l'lment de rfrence  sa place
    int tmp  = tab[ref];
    tab[ref] = tab[mid];
    tab[mid] = tmp;

    ret* r = malloc(sizeof(*r));
    r->low = low;
    r->mid = mid;

    // Retourner l'indice correct de l'lment de rfrence
    return r;
}

/* Cette fonction trie la partie [low:high] du tableau tab par des appels 
 * rcursifs  elle-mme et  la fonction de partition. */
static void quicksort(int* tab, int low, int high) {
    if (low < high) {
        /* Evidemment, on ne va pas s'amuser  trier la partie qu'on sait qui 
         * est gale  la rfrence. Cette partie s'tend sur tab[low:mid-1] 
         * si low et mid sont ceux trouvs en fin de fonction partition(). */
        ret* data = partition(tab, low, high);
        quicksort(tab, low, data->low - 1); // Trier partie < ref
        quicksort(tab, data->mid, high);    // Trier partie > ref
                                            // Ignorer partie == ref

        free(data);
    }
}

int main() {
    const int N = 20;


    printf("This program sorts lists of %d positive or null integers, from "
           "0 to 99, using 3-way Quicksort.\nIt was NOT made to be robust, "
           "wrong inputs aren't managed!\n", N);
    while (true) {
        int seed = 0;
        printf("\nNew test -- Enter random seed (integer): ");
        scanf("%d", &seed);

        
        srand(seed);
        int a[N];
        for (int i = 0; i < N; i++)
            a[i] = rand() % 100;
    
        printArray(a, N);
        quicksort(a, 0, N-1);
        printArray(a, N);
    }

    return 0;
}
