#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

static void printArray(int* tab, int n) {
    printf("Actual state:\n");
    for (int i = 0; i < n; i++)
        printf("%d ", tab[i]);
    
    printf("\n");
}


static int partition(int* tab, int low, int high) {
    // Partitionnement 2-way
    int ref = high--;

    /* Cette technique rorganise le tableau de la manire suivante :
     On observe les lments compris entre 0 et high - 1.
     On va diviser le tableau tab[low, ..., ref-1] en 2 parties :  gauche,
     tous les lments <=  tab[ref] et ensuite tous les lments
     strictement suprieurs  tab[ref]. Quand cette action est effectue,
     il ne reste plus qu' changer tab[ref] et le tab[low final].

     Comment se passe le partitionnement ? A chaque tour de boucle,
     l'invariant suivant est respect :
         tab[0 : low-1]          : <= tab[ref] (partie A)
         tab[low : high]         : indtermin,  analyser (partie B)
         tab[high + 1 : ref - 1] : > tab[ref] (partie C)
         La partie B se rduit d'une unit  chaque itration
     
     La zone indtermine est rduite  chaque tour de boucle, que ce soit par
     la droite (dcrment du marqueur high) ou par la gauche (incrment du
     marqueur low). A chaque tour, on regarde si tab[low] est plus petit ou
     gal  la rfrence.
     Si il est infrieur, on n'y touche pas et on
     incrmente low (en effet, vu l'invariant, cela veut dire que l'lment
     est  sa place : on largit la partie A et on rduit B).
     Si il est suprieur, on le switche avec l'lment  l'indice high (on
     ne sait pas ce que celui-ci vaut, btw), on dcrmente ensuite high sans 
     toucher  low qui au prochain tour de boucle, permettera d'valuer cet
     lment switch inconnu et le placer  la bonne place. On sait que l'
     invariant reste respect : C est largie, alors que B est rduite, et l'
     lment est bien plac selon l'invariant.

     A la fin de l'algorithme,  la sortie de la boucle, on place l'lment de
     rfrence au bon endroit (on le switch avec tab[low]). Pourquoi low ?
     Parce que, que l'on termine la boucle par un high-- ou un low++, low
     pointera vers le premier lment de tab qui est strictement plus grand
     que tab[ref]. Lors de l'avant-dernire itration, on a high - low = 1.
     Ensuite, qu'on excute high-- ou low++, on arrive  low = high. Le dernier
     tour de boucle compare donc tab[low=high] avec lui-mme, on incrmente
     donc low une dernire fois. Dsormais, low - high = 1, la boucle s'arrte,
     et on a plus qu' changer tab[ref] avec tab[low] pour placer tab[ref] au
     bon endroit.
     La postcondition est respecte,  savoir, on a dsormais 2 parties (dont
     l'une peut tre vide) spares par l'lment de rfrence : avant lui,
     les lments ont une valeurs inf. ou gale, et aprs lui sont strict. sup.
    */
    while (low <= high) {
        if (tab[low] <= tab[ref])
            low++;
        else {
            int tmp   = tab[high];
            tab[high] = tab[low];
            tab[low]  = tmp;
            high--;
        }
    }

    int tmp  = tab[ref];
    tab[ref] = tab[low];
    tab[low] = tmp;
    return low;
}

static void quicksort(int* tab, int low, int high) {
    if (low < high) {
        int q = partition(tab, low, high);
        quicksort(tab, low, q-1);
        quicksort(tab, q, high);
    }
}

int main() {
    const int N = 20;

    while (true) {
        int seed = 0;
        printf("\nEnter seed: ");
        scanf("%d", &seed);

        
        srand(seed);
        int a[N];
        for (int i = 0; i < N; i++)
            a[i] = rand() % 100;
    
        a[5] = a[6] = a[7] = a[N-1]; //Force randomness to see dutch flag effect
    
        printArray(a, N);

        quicksort(a, 0, N-1);

        printArray(a, N);
    }

    return 0;
}